////////////////////////////////////////////////////////////
// ATL Sort.cpp - Sample project featuring sorting templates for ATL
//
// Copyright (C) Alax.Info, 2006
// http://alax.info
//
// A permission to use the source code is granted as long as reference to 
// source website http://alax.info is retained.

#include "stdafx.h"
#include <atlcoll.h>

#pragma warning(disable: 4244) // VC++.NET unexpectedly barfs with this warning, see **1 below

////////////////////////////////////////////////////////////
// CSimpleSortTraitsT

template <typename _Element, typename _Parameter = INT_PTR>
class CSimpleSortTraitsT
{
public:
	typedef typename _Element ELEMENT;
	typedef typename _Element ELEMENTARGUMENT;
	typedef typename _Parameter PARAMETER;
	typedef typename _Parameter PARAMETERARGUMENT;

public:
// CSimpleSortTraitsT
	static PARAMETER GetDefaultParameter() throw()
	{
		return (PARAMETER) 0;
	}
	static INT CompareElements(ELEMENTARGUMENT Element1, ELEMENTARGUMENT Element2, PARAMETERARGUMENT Parameter) throw()
	{
		Parameter;
		return (Element1 < Element2) ? -1 : (Element1 > Element2) ? 1 : 0;
	}
	static VOID SwapElements(ELEMENT& Element1, ELEMENT& Element2, PARAMETERARGUMENT Parameter) throw()
	{
		Parameter;
		ELEMENT Buffer = Element1;
		Element1 = Element2;
		Element2 = Buffer;
	}
};

////////////////////////////////////////////////////////////
// CDefaultSortTraitsT

template <typename _Element, typename _Parameter = INT_PTR>
class CDefaultSortTraitsT
{
public:
	typedef typename _Element ELEMENT;
	typedef typename const _Element& ELEMENTARGUMENT;
	typedef typename _Parameter PARAMETER;
	typedef typename _Parameter PARAMETERARGUMENT;

public:
// CDefaultSortTraitsT
	static PARAMETER GetDefaultParameter() throw()
	{
		return (PARAMETER) 0;
	}
	static INT CompareElements(ELEMENTARGUMENT Element1, ELEMENTARGUMENT Element2, PARAMETERARGUMENT Parameter) throw()
	{
		Parameter;
		return (Element1 < Element2) ? -1 : (Element1 > Element2) ? 1 : 0;
	}
	static VOID SwapElements(ELEMENT& Element1, ELEMENT& Element2, PARAMETERARGUMENT Parameter) throw()
	{
		Parameter;
		BYTE* pnBuffer = (BYTE*) alloca(sizeof (ELEMENT));
		CopyMemory(pnBuffer, &Element1, sizeof (ELEMENT));
		CopyMemory(&Element1, &Element2, sizeof (ELEMENT));
		CopyMemory(&Element2, pnBuffer, sizeof (ELEMENT));
	}
};

////////////////////////////////////////////////////////////
// _SortHelper

class _SortHelper
{
public:
// _SortHelper
	static inline BOOL Ascending(INT nCompareResult) throw()
	{
		return nCompareResult < 0;
	}
	static inline BOOL Descending(INT nCompareResult) throw()
	{
		return nCompareResult > 0;
	}
	template <typename _Traits>
	static VOID BubbleSort(typename _Traits::ELEMENT* pElements, SSIZE_T nFirstIndex, SSIZE_T nLastIndex, typename _Traits::PARAMETERARGUMENT Parameter = _Traits::GetDefaultParameter(), BOOL (*CompareResultTraitFunction)(INT) = Ascending) throw()
	{
		for(SSIZE_T nIndex1 = nLastIndex; nIndex1 > nFirstIndex; nIndex1--)
			for(SSIZE_T nIndex2 = nFirstIndex; nIndex2 < nIndex1; nIndex2++)
				if(!CompareResultTraitFunction(_Traits::CompareElements(pElements[nIndex2], pElements[nIndex2 + 1], Parameter)))
					_Traits::SwapElements(pElements[nIndex2], pElements[nIndex2 + 1], Parameter);
	}
	template <typename _Traits>
	static VOID QuickSort(typename _Traits::ELEMENT* pElements, SSIZE_T nFirstIndex, SSIZE_T nLastIndex, typename _Traits::PARAMETERARGUMENT Parameter = _Traits::GetDefaultParameter(), BOOL (*CompareResultTraitFunction)(INT) = Ascending) throw()
	{
		if(nLastIndex - nFirstIndex < 1)
			return;
		SSIZE_T nIndex1 = nFirstIndex, nIndex2 = nLastIndex;
		SSIZE_T nIndex3 = (nIndex1 + nIndex2) / 2;
		while(nIndex1 <= nIndex2)
		{
			while(CompareResultTraitFunction(_Traits::CompareElements(pElements[nIndex1], pElements[nIndex3], Parameter)))
				nIndex1++;
			while(CompareResultTraitFunction(_Traits::CompareElements(pElements[nIndex3], pElements[nIndex2], Parameter)))
				nIndex2--;
			if(nIndex1 <= nIndex2)
			{
				if(nIndex3 == nIndex1)
					nIndex3 = nIndex2;
				else if(nIndex3 == nIndex2)
					nIndex3 = nIndex1;
				_Traits::SwapElements(pElements[nIndex1++], pElements[nIndex2--], Parameter);
			}
		}
		if(nFirstIndex < nIndex2)
			QuickSort<_Traits>(pElements, nFirstIndex, nIndex2, Parameter, CompareResultTraitFunction);
		if(nIndex1 < nLastIndex)
			QuickSort<_Traits>(pElements, nIndex1, nLastIndex, Parameter, CompareResultTraitFunction);
	}
};

////////////////////////////////////////////////////////////
// Helpers

static UINT g_nRandomSeed = 0;

// Random function returns pseudo-random value in range [0, nRange - 1]

__declspec(noinline) UINT Random(UINT nRange) throw()
{
	ATLASSERT(nRange > 0);
	g_nRandomSeed = (g_nRandomSeed + 0x23456789) * 7;
	return g_nRandomSeed % nRange;
}

////////////////////////////////////////////////////////////
// Main

class CComplexValue
{
public:
	INT m_nA;
	INT m_nB;

public:
// CComplexValue
	CComplexValue() throw()
	{
	}
	CComplexValue(INT nA, INT nB) throw() :
		m_nA(nA),
		m_nB(nB)
	{
	}
};

class CComplexValueSortTraits :
	public CDefaultSortTraitsT<CComplexValue>
{
public:
// CComplexValueSortTraits
	static INT CompareElements(const CComplexValue& Element1, const CComplexValue& Element2, PARAMETERARGUMENT) throw()
	{
		INT nC1 = Element1.m_nA * Element1.m_nB;
		INT nC2 = Element2.m_nA * Element2.m_nB;
		return nC1 - nC2;
	}
};

class CComplexValueIndexSortTraits :
	public CSimpleSortTraitsT<SIZE_T, const CAtlArray<CComplexValue>*>
{
public:
// CComplexValueSortTraits
	static INT CompareElements(SIZE_T nIndex1, SIZE_T nIndex2, const CAtlArray<CComplexValue>* pValues) throw()
	{
		const CComplexValue& Element1 = pValues->GetAt(nIndex1);
		const CComplexValue& Element2 = pValues->GetAt(nIndex2);
		INT nC1 = Element1.m_nA * Element1.m_nB;
		INT nC2 = Element2.m_nA * Element2.m_nB;
		return nC1 - nC2;
	}
};

int _tmain(int argc, _TCHAR* argv[])
{
	// 1. Sorting INT values, using default traits class; using "Descending" order
	{
		printf("Sorting array of INT's:\n");
		CAtlArray<INT> Values;
		// Populate array
		{
			printf("Original values:\n");
			enum { COUNT = 15, RANGE = 100 };
			for(SIZE_T nIndex = 0; nIndex < COUNT; nIndex++)
			{
				INT nValue = Random(RANGE);
				printf("\t%5d\n", nValue);
				ATLVERIFY(Values.Add(nValue) >= 0);
			}
			printf("\n");
		}
		typedef CSimpleSortTraitsT<INT> CIntSortTraits;
		if(!Values.IsEmpty())
			_SortHelper::QuickSort<CIntSortTraits>(Values.GetData(), 0, Values.GetCount() - 1, 0, _SortHelper::Descending);
		{
			printf("Sorted values:\n");
			for(SIZE_T nIndex = 0; nIndex < Values.GetCount(); nIndex++)
				printf("\t%5d\n", Values[nIndex]);
			printf("\n");
		}
		printf("\n");
	}
	// 2. Sorting CComplexValue values, using custom traits class to sort by expression
	{
		printf("Sorting array of CComplexValue's:\n");
		CAtlArray<CComplexValue> Values;
		// Populate array
		{
			printf("Original values:\n");
			enum { COUNT = 15, RANGE = 100 };
			for(SIZE_T nIndex = 0; nIndex < COUNT; nIndex++)
			{
				CComplexValue Value(Random(RANGE * 2 + 1) - RANGE, Random(RANGE * 2 + 1) - RANGE);
				printf("\t%5d\t%5d\t%5d\n", Value.m_nA, Value.m_nB, Value.m_nA * Value.m_nB);
				ATLVERIFY(Values.Add(Value) >= 0);
			}
			printf("\n");
		}
		if(!Values.IsEmpty())
			_SortHelper::QuickSort<CComplexValueSortTraits>(Values.GetData(), 0, Values.GetCount() - 1, 0);
		{
			printf("Sorted values (see third column):\n");
			for(SIZE_T nIndex = 0; nIndex < Values.GetCount(); nIndex++)
				printf("\t%5d\t%5d\t%5d\n", Values[nIndex].m_nA, Values[nIndex].m_nB, Values[nIndex].m_nA * Values[nIndex].m_nB);
			printf("\n");
		}
		printf("\n");
	}
	// 3. Sorting indexes into CComplexValue value array, using custom traits class to sort by expression
	{
		printf("Sorting array of CComplexValue indexes:\n");
		CAtlArray<CComplexValue> Values;
		CAtlArray<SIZE_T> Indexes;
		// Populate array
		{
			printf("Original values:\n");
			enum { COUNT = 15, RANGE = 100 };
			for(SIZE_T nIndex = 0; nIndex < COUNT; nIndex++)
			{
				CComplexValue Value(Random(RANGE * 2 + 1) - RANGE, Random(RANGE * 2 + 1) - RANGE);
				printf("\t%5d\t%5d\t%5d\n", Value.m_nA, Value.m_nB, Value.m_nA * Value.m_nB);
				ATLVERIFY(Values.Add(Value) >= 0);
				ATLVERIFY(Indexes.Add(nIndex) >= 0); // **1 - see above
			}
			printf("\n");
		}
		if(!Values.IsEmpty())
			_SortHelper::QuickSort<CComplexValueIndexSortTraits>(Indexes.GetData(), 0, Indexes.GetCount() - 1, &Values);
		{
			printf("Sorted values (see third column):\n");
			for(SIZE_T nIndex = 0; nIndex < Indexes.GetCount(); nIndex++)
				printf("\t%5d\t%5d\t%5d\n", Values[Indexes[nIndex]].m_nA, Values[Indexes[nIndex]].m_nB, Values[Indexes[nIndex]].m_nA * Values[Indexes[nIndex]].m_nB);
			printf("\n");
		}
		printf("\n");
	}
	return 0;
}

